//
//  main.c
//
//  Mini watt hour meter 0.4c
//  Copyright (C) 2009 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Non committal 7channel watt hour meter.  
//		Use for wall outlet power checker.
//		Voltage is 100V(RMS) fixed.
//
//	Device:	PIC16F88
//	Clock:	4MHz/Internal
//	Internal peripheral module:
//		WDT:	1sec
//		BOR:	-
//		EEPROM:	Use
//		Timer0:	-
//		Timer1:	2Hz
//		Timer2:	-
//		CCP:	-
//		SSP:	-
//		AUSART:	-
//		ADC:	Current measure
//		Comp:	-
//		Vref:	-
//	Pin assign:		Refer config.h
//	Charactor LCD:	SC1602
//	CT:				U_RD Co.,Ltd  CTL-10-CLS
//	RL:				500ohm -> Max20A
//
//	Environment:
//		IDE: MPLAB IDE 8.36
//      HI-TECH C PRO for the PIC10/12/16 MCU Family(Lite) V9.65PL1
//
//
//	Update history
//	---------- ----- -------------------------------------------
//	2009.10.30 v0.4  PIC16/HI-TECH C version
//	2009.10.31 v0.4  Time slice method (Current measure)
//	2009.11.04 v0.4a Max10A->20A RL=1k->500ohm
//	2009.11.04 v0.4a ADC clock 1/32->1/16
//	2009.11.04 v0.4a Round 4/5
//	2009.11.04 v0.4a LCD initial wait
//	2009.11.04 v0.4a Adjust Timer1 period
//	2009.11.04 v0.4a Bug fix (Recall data is little)
//	2009.11.11 v0.4b RL -> config.h
//	2009.11.22 v0.4c Bug fix, display overflow -32kWh
//	2009.12.15 v0.4d Bug fix, No clear of month
//

#include <stdio.h>

#include "htc.h"						// HI-TECH C
#include "eeprom_routines.h"			// HI-TECH C

#include "config.h"						// Hardware configuration
#include "lcd.h"						// Character LCD support


#define TMR1PERIOD		62250			// 4MHz/4/8/62500=2Hz  CHG 2009/11/22 ISR=-250
#define CH_COUNT		7				// CT number
#define SAMP_COUNT		1000			// ADC
#define SW1_LONGPUSH	100
#define SW2_LONGPUSH	100
#define YEN_RATE		25				// 1kWh
#define EE_SIZE			256				// Boundery sizeof(ee_buf)
#define WAKEMSG							// ADD 2009.11.22  Option

#define DISP_MODE_NORMAL	1
#define DISP_MODE_HIST		2


//
//	Config
//
//	CPU Configuration Fuses
__CONFIG(INTIO & MCLRDIS & LVPDIS);

//	EEPROM initial data
// YY,MM,CH0-Lowbyte,CH0-Highbyte,...
__EEPROM_DATA(9, 1, 0, 0, 0, 1, 0, 2); __EEPROM_DATA(0, 3, 0, 4, 0, 5, 0, 6);
__EEPROM_DATA(9, 2, 0, 0, 1, 0, 2, 0); __EEPROM_DATA(3, 0, 4, 0, 5, 0, 6, 0);
__EEPROM_DATA(9, 3, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 4, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 5, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 6, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 7, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 8, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9, 9, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9,10, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(9,11, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);	// UPD 2009.11.22
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);
__EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0); __EEPROM_DATA(0, 0, 0, 0, 0, 0, 0, 0);

//
// Global area
//
// DateTime
static char ddom[] = {0, 31,28,31,30,31,30,31,31,30,31,30,31};
static char yy;
static char mm;
static char dd = 1;
static char hh = 0;
static char mi = 0;
static char ss = 0;
static char tmr1flg = 0;
// ADC
static long adcval[CH_COUNT] = {-1, -2, -3, -4, -5, -6, -7};
static char adch  = 0;
static long adsum = 0;
static int  adcnt = SAMP_COUNT;


//
//	EEPROM I/O
//
void eeprom_writes(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
		EEPROM_WRITE(addr++, *buf++);
	}
	return;
}
void eeprom_reads(unsigned char addr, unsigned char *buf, unsigned char len){
	for(; len > 0; --len) {
		*buf++ = EEPROM_READ(addr++);
	}
	return;
}

//
// Override stdio
//
void putch(char c) {
	LCD_WriteData(c);
}

//
// Intrrupt handler
//
void interrupt isr(void) {

	// ADC
	if (ADIE && ADIF) {
		adsum += ((ADRESH << 8) | ADRESL);
		if (--adcnt < 0) {
			adcnt = SAMP_COUNT;
			adcval[adch] = adsum;
			adsum = 0;
			if (++adch >= CH_COUNT) {
				adch = 0;
			}
			ADCON0 = (ADCON0 & 0b11000111) | (adch << 3);
		}
		ADIF = 0;
		ADGO = 1;						// Kick next measure
	}

	// Timer1 (2Hz)
	if (TMR1IE && TMR1IF) {
		TMR1H = (-TMR1PERIOD) >> 8;		// Preset
		TMR1L = (-TMR1PERIOD) & 0xff;
		TMR1IF=0;

		// DateTime count up
		if ((tmr1flg++) & 1) {			// 1Hz
			ss++;
			if (ss > 59) {
				ss = 0;
				mi++;
				if (mi > 59) {
					mi = 0;
					hh++;
					if (hh > 23) {
						hh = 0;
						if (((yy % 4) == 0) && ((yy % 100) != 0)) {
							ddom[2] = 29;
						} else {
							ddom[2] = 28;
						}
						dd++;
						if (dd > ddom[mm]) {
							dd = 1;
							mm++;
							if (mm > 12) {
								mm = 1;
								yy++;
							}
						}
					}
				}
			}
		}
	}

	return;
}

//
// Date Time key input & modify
//
void dateset(void) {
	unsigned char y,m,d,h,n;
	unsigned char p;

	y = yy;
	m = mm;
	d = dd;
	h = hh;
	n = mi;
	p = 0;
	LCD_Control(LCD_CURSOR_BLINK);

	for(;;) {
		const unsigned char col[] = {3, 6, 9, 12, 15};

		// display
		LCD_Position(0, 0);
		printf("%04d.%02d.%02d %02d:%02d", y+2000, m, d, h, n);
		LCD_Position(0, col[p]);
		
		// Key off wait
		while (SW1 || SW2) {
			CLRWDT();
		}

		// Key push wait
		while (!SW1 && !SW2) {
			CLRWDT();
		}

		// Key push
		if (SW1) {
			p++;
			if (p > 4) break;			// forever exit
		}
		if (SW2) {
			switch(p) {
			case 0:
				y++;
				if (y > 99) {
					y = 0;
				}
				break;
			case 1:
				m++;
				if (m > 12) {
					m = 1;
				}
				break;
			case 2:
				d++;
				if (d > ddom[m]) {
					d = 1;
				}
				break;
			case 3:
				h++;
				if (h > 23) {
					h = 0;
				}
				break;
			default:
				n++;
				if (n > 59) {
					n = 0;
				}
				break;
			}
		}

		// Chattering cancel
//		__delay_ms(100);
	}
	
	// Fix
	yy = y;
	mm = m;
	dd = d;
	hh = h;
	mi = n;
	ss = 0;
	LCD_Control(LCD_DISP_ON);

	// Key off wait
	while (SW1 || SW2) {
		CLRWDT();
	}

	return;
}

//
//  M a i n
//
main(){
	static int  w[CH_COUNT]  = {0, 0, 0, 0, 0, 0, 0};
	static long ws[CH_COUNT] = {0, 0, 0, 0, 0, 0, 0};

	unsigned char i;					// Multi purpose, short term
	unsigned char ct;
	unsigned char b_ss;
	unsigned char b_dd;
	unsigned char b_mm;
	static unsigned char sw1_scan;
	static unsigned char sw2_scan;
	static unsigned char disp_ch;
	static unsigned char disp_mode;
	static unsigned char disp_ee;
	static unsigned char disp_ee_b;
	static unsigned char ee_addr;
	static struct {
		char yy;
		char mm;
		int wh[CH_COUNT];
	} ee_buf;


//		*** System clock ***
//
//				  OSCCON - Oscillator control register
//				  -0000000  Initial
//				  .XXX....  IRCF: Internal oscillator
//				  .000....     31.25KHz
//				  .001....     125KHz
//				  .010....     250KHz
//				  .011....     500KHz
//				  .100....     1MHz
//				  .101....     2MHz
//				  .110....     4MHz
//				  .111....     8MHz
//				  ....X...  OSTS: Oscillator Start-up Time-out Status bit
//				  ....0...     Device is running from T1OSC or INTRC as secondary system clock
//				  ....1...     Device is running from the primery system clock
//				  .....X..  IOFS:INTOSC Frequency stable bit
//				  .....1..     stable
//				  .....0..     not stable
//				  ......XX  SCS: Oscillator Mode Select bits
//				  ......00    Oscillator mode defined by FOSC<2:0>
//				  ......01    T1OSC is used by system clock
//				  ......10    Internal RC is used for system clock
//				  ......11    Reserved
	OSCCON =    0b01100000;		// Internal 4MHz

//				  OSCTUNE - Oscillator tuning register
//				  00000000  Initial
//				  ..XXXXXX  TUN: Frequency tuning bits
//				  ..011111     +31	+12.5%
//				  ..000001     +1	+0.4%
//				  ..000000     0
//				  ..111111     -1	-0.4%
//				  ..100000     -32	-12.5%
	OSCTUNE = 0;

//		*** System option ***
//
//				  OPTION_REG - Option register
//				  11111111  Initial
//				  X.......  RBPU: PORTB Pull-up enable
//				  1.......     disabled
//				  0.......     enable
//				  .X......  INTEDG: Interrupt edge select
//				  .1......     riging edge
//				  .0......     failling edge
//				  ..X.....  T0CS: TMR0 clock source select
//				  ..1.....     Transition on GP2/T0CKI pin
//				  ..0.....     Internal instruction cycle
//				  ...X....  T0SE: TMR0 source edge select
//				  ...1....     high-to-low edge
//				  ...0....     low-to-high edge
//				  ....X...  PSA: Prescaler assignment
//				  ....1...     WDT
//				  ....0...     Timer0
//				  .....XXX  PS:  Prescaler assignment
//				  .....000     TMR0 rate 1:2 / WDT rate 1:1
//				  .....001               1:4 /          1:2
//				  .....010               1:8 /          1:4
//				  .....011              1:16 /          1:8
//				  .....100              1:32 /         1:16
//				  .....101              1:64 /         1:32
//				  .....110             1:128 /         1:64
//				  .....111             1:256 /        1:128
	OPTION =    0b11010100;		// Timer0=CLKO(4MHz/4)/32 -> 31.25KHz

//		*** Watchdock timer ***
//
//				  WDTCON - Watchdog control register
//				  ---01000  Initial
//				  ...XXXX.  WDTPS: Watchdog timer period select bits
//				  ...0000.     1:32
//				  ...0001.     1:64
//				  ...0010.     1:128
//				  ...0011.     1:256
//				  ...0100.     1:512
//				  ...0101.     1:1024
//				  ...0110.     1:2048
//				  ...0111.     1:4096
//				  ...1000.     1:8192
//				  ...1001.     1:16384
//				  ...1010.     1:32768
//				  ...1011.     1:65536
//				  .......X  WDTPS: Watchdog timer period select bits
//				  .......1     WDT is turned on
//				  .......0     WDT is turned off
	WDTCON =    0b00010101;		// 32.25KHz / 32.768KHz ~= 1sec  <disable>

//		*** GPIO port ***
//
//				  PORTA - General purpose I/O register A
//                xxx00000  Initial
//				  X.......  RA7: OSC1/CLKI
//				  .X......  RA6: OSC2/CLKO
//				  ..X.....  RA5: MCLR/Vpp
//				  ...X....  RA4: AN4/T0CKI
//				  ....X...  RA3: AN3/Vref+
//				  .....X..  RA2: AN2/Vref-
//				  ......X.  RA1: AN1
//				  .......X  RA0: AN0
	PORTA =     0b00000000;		// Initial all off

//				  TRISA - GPIO tri-state register A
//                11111111  Initial
//				  1.......  TRISA7: RA7 1=input only
//				  .0......  TRISA6: RA6 0=output only
//				  ..1.....  TRISA5: RA5 1=input only
//				  ...X....  TRISA4: RA4 1=input 0=output
//				  ....X...  TRISA3: RA3 1=input 0=output
//				  .....X..  TRISA2: RA2 1=input 0=output
//				  ......X.  TRISA1: RA1 1=input 0=output
//				  .......X  TRISA0: RA0 1=input 0=output
	TRISA =     0b10111111;		// RA6=output

//				  PORTB - General purpose I/O register B
//                xxxxxxxx  Initial
//				  X.......  RB7: AN6/T1OSI/PGD
//				  .X......  RB6: AN5/T1OSO/T1CKI/PGC
//				  ..X.....  RB5: SS/TX/CK
//				  ...X....  RB4: SCK/SCL
//				  ....X...  RB3: CCP1/PGM
//				  .....X..  RB2: SDO/RX/DT
//				  ......X.  RB1: SDI/SDA
//				  .......X  RB0: INT/CCP
	PORTB =     0b00000000;		// Initial all off

//				  TRISB - GPIO tri-state register B
//                11111111  Initial
//				  X.......  TRISB7: RB7 1=input 0=output
//				  .X......  TRISB6: RB6 1=input 0=output
//				  ..X.....  TRISB5: RB5 1=input 0=output
//				  ...X....  TRISB4: RB4 1=input 0=output
//				  ....X...  TRISB3: RB3 1=input 0=output
//				  .....X..  TRISB2: RB2 1=input 0=output
//				  ......X.  TRISB1: RB1 1=input 0=output
//				  .......X  TRISB0: RB0 1=input 0=output
	TRISB =     0b11000000;		// RB4,5=output

//		*** Timer 0 ***
//
//				  TMR0 - Timer0 module register
//				  ????????  Initial
	TMR0 = 		0;				// 32.25KHz / 8bit = max122Hz

//		*** Timer 1 ***
//
//				  T1CON - Timer1 control register
//				  -0000000  Initial
//				  .X......  T1RUN: Timer1 System Clock Status bit (R/O)
//				  .1......    System clock is derived from Timer1 oscillator
//				  .0......    System clock is derived from another source
//				  ..XX....  T1CKPS: Timer1 Input Clock Prescale Select bits
//				  ..11....    1:8 Prescale value
//				  ..10....    1:4 Prescale value
//				  ..01....    1:2 Prescale value
//				  ..00....    1:1 Prescale value
//				  ....X...  T1OSCEN: Timer1 Oscillator Enable Control bit
//				  ....1...    Oscillator is enabled
//				  ....0...    Oscillator is shut off (the oscillator inverter is turned off to eliminate power drain)
//				  .....X..  T1SYNC:Timer1 External Clock Input Synchronization Control bit
//				  .....1..    TMR1CS = 1: Do not synchronize external clock input
//				  .....0..                Synchronize external clock input
//				              TMR1CS = 0: This bit is ignored. Timer1 uses the internal clock when TMR1CS = 0
//				  ......X.  TMR1CS: Timer1 Clock Source Select bit
//				  ......1.    External clock from pin RB6/AN5(1)/PGC/T1OSO/T1CKI (on the rising edge)
//				  ......0.    Internal clock (FOSC/4)
//				  .......X  TMR1ON: Timer1 On bit
//				  .......1    Enables Timer1
//				  .......0    Stops Timer1
	T1CON =     0b00110001;

//		*** Timer 2 ***
//

//		*** A/D Converter ***
//
//				  ADCON0 - A/D control register 0
//				  00000000  Initial
//				  XX......  ADCS: A/D converter clock select
//				  00......    ADCS2=0 Fosc/2   ADCS2=1 Fosc/4
//				  01......            Fosc/8           Fosc/16
//				  10......            Fosc/32          Fosc/64
//				  11......            FRC              FRC
//				  ..XXX...  CHS: Analog channel select bits
//				  ..000...    Channel 0 RA0/AN0
//				  ..001...    Channel 1 RA1/AN1
//				  ..010...    Channel 2 RA2/AN2
//				  ..011...    Channel 3 RA3/AN3
//				  ..100...    Channel 4 RA4/AN4
//				  ..101...    Channel 5 RB6/AN5
//				  ..110...    Channel 6 RB7/AN6
//				  .....X..  GO/DONE: A/D conversion status bit
//				  .....1..    ADON=1 in progress
//				  .....0..           not in progress
//				  .......X  ADON: A/D on bit
//				  .......1    module is operating
//				  .......0    module is shutoff
//	ADCON0 =    0b10000001;		// Fosc/2 <Channel dynamic change>  DEL 2009.11.04
	ADCON0 =    0b01000001;		// Fosc/8 <Channel dynamic change>  ADD 2009.11.04

//				  ADCON1 - A/D control register 1
//				  00000000  Initial
//				  X.......  ADFM: A/D result format select bit
//				  1.......    Right justified
//				  0.......    Left justified
//				  .X......  ADCS2: A/D clock divied by 2 select bit
//				  .1......    Enabled
//				  .0......    Disabled
//				  ..XX....  VCFG: A/D port configuration control bits
//				  ..00....    VREF+=AVdd  VREF-=AVss
//				  ..01....    VREF+=AVdd  VREF-=VREF-
//				  ..10....    VREF+=VREF+ VREF-=AVss
//				  ..11....    VREF+=VREF+ VREF-=VREF-
//	ADCON1 =    0b10000000;		// Right adj, clock/1, Vref=Vdd:Vss  DEL 2009.11.04
	ADCON1 =    0b11000000;		// Right adj, clock/2, Vref=Vdd:Vss  ADD 2009.11.04

//				  ANSEL - Analog select register
//				  01111111  Initial
//				  .X......  ANS6 1=Analog 0=Digital
//				  ..X.....  ANS5 1=Analog 0=Digital
//				  ...X....  ANS4 1=Analog 0=Digital
//				  ....X...  ANS3 1=Analog 0=Digital
//				  .....X..  ANS2 1=Analog 0=Digital
//				  ......X.  ANS1 1=Analog 0=Digital
//				  .......X  ANS0 1=Analog 0=Digital
	ANSEL =     0b01111111;		// All analog

//		*** Interrupt ***
//
//				  PIE1 - Peripheral interrupt enable register 1
//                -0000000  Initial
//							0=disable 1=enable
//				  .X......  ADIE  : A/D Converter interrupt
//				  ..X.....  RCIE  : AUSART Receive interrupt
//				  ...X....  TXIE  : AUSART Transmit interrupt
//				  ....X...  SSPIE : SSP interrupt
//				  .....X..  CCP1IE: CCP1 interrupt
//				  ......X.  TMR2IE: Timer2 to PR2 match overflow
//				  .......X  TMR1IE: Timer1 overflow
	PIE1 =      0b01000001;		// ADC+TMR1 enable

//				  PIE2 - Peripheral interrupt enable register 2
//                -0000000  Initial
//							0=disable 1=enable
//				  X.......  OSFIE : Oscillator fail
//				  .X......  CMIE  : Comparator interrupt
//				  ...X....  EEIE  : EEPROM Write complete
	PIE2 =      0b00000000;		// Off

//				  INTCON - Interrupt control register
//                 00000000  Initial
//							0=disable 1=enable
//				  X.......  GIE : Global interrupt
//				  .X......  PEIE: Peripheral intterupt -> PIE1
//				  ..X.....  T0IE: TMR0 Overflow interrupt
//				  ...X....  INTE: RB0/INT external interrupt
//				  ....X...  RBIE: RB Port Change interrupt
//							0=not(Appl reset) 1=occurred
//				  .....X..  T0IF: TMR0 Overflow interrupt flag
//				  ......X.  INTF: RB0/INT external interrupt flag
//				  .......X  GPIF: RB Port Change interrupt flag
	INTCON =    0b11000000;		// PIE1 enable

//		*** LCD module ***
	LCD_Init();							// ADD 2009.11.22
	LCD_Start();						// ADD 2009.11.22


	//
	// Wakeup message
	//
#ifdef WAKEMSG							// ADD 2009.11.22
	{
		for(i = 0; i < 10; i++) {		// ADD 2009.11.04
			__delay_ms(100);			// ADD 2009.11.04 
			CLRWDT();					// ADD 2009.11.04
		}								// ADD 2009.11.04
//		LCD_Init();						// DEL 2009.11.22
//		LCD_Start();					// DEL 2009.11.22
		puts("WHM0.4d 09.12.15");		// UPD 2009.12.15
		for(i = 0; i < 10; i++) {
			__delay_ms(100);
			CLRWDT();
		}
	}
#endif									// ADD 2009.11.22

	//
	// Recall variable
	//
	{
		unsigned char p;

		// Search EEPROM data
		p = 0;
		for(i = 0; i < (EE_SIZE/sizeof(ee_buf)); i++) {
			eeprom_reads((unsigned char)p, (char *)&ee_buf, 2);
			if ((yy < ee_buf.yy) && (mm < ee_buf.mm)) {
				ee_addr = p;
			}
			p += sizeof(ee_buf);
		}

		// Get data
		eeprom_reads(ee_addr, (char *)&ee_buf, sizeof(ee_buf));
		yy = ee_buf.yy;
		mm = ee_buf.mm;
		for(i = 0; i < CH_COUNT; i++) {
//			ws[i] = ee_buf.wh[i] * 3600;		// DEL 2009.11.04
			ws[i] = ee_buf.wh[i] * 3600000L;	// ADD 2009.11.04  Ws <- kWh
		}
	}

	//
	// Initial data
	//
	b_ss      = ss;
	b_dd      = dd;
	b_mm      = mm;
	sw1_scan  = 0;
	sw2_scan  = 0;
	disp_ch   = 0;
	disp_mode = DISP_MODE_NORMAL;
	disp_ee   = ee_addr;
	disp_ee_b = 1;

	//
	// Evant loop
	//
	ADGO = 1;							// ADC Kick start
	for(ct = 0; ; ct++){

		//
		// Second task
		//
		if (b_ss != ss) {
			b_ss = ss;

			// Accumerate watt*sec
			for(i = 0; i < CH_COUNT; i++) {
				// [W] = (sum(ADC) / SAMP_COUNT * 2) * (Vdd/resolution) * CT_Rate * Vol
				//         adcval  /    1000    * 2  *  5.0  / 1024     *    K    * 100
				w[i] = (adcval[i] * RL_Kx10 + 5120L) / 10240L;		// CHG 2009.11.11
				ws[i] += w[i];
			}
		}

		//
		// Daily task
		//
		if (b_dd != dd) {
			b_dd = dd;

			// Snap shot to EEPROM
			ee_buf.yy = yy;
			ee_buf.mm = mm;
			for(i = 0; i < CH_COUNT; i++) {
				ee_buf.wh[i] = ws[i] / 3600000;		// Ws -> kWh
			}
			eeprom_writes(ee_addr, (char *)&ee_buf, sizeof(ee_buf));
		}

		//
		// Monthly task
		//
		if (b_mm != mm) {
			b_mm = mm;

			// Save addr update EEPROM
			ee_addr += sizeof(ee_buf);	// No overflow

			// Clear Wh						// ADD 2009.12.15
			for(i = 0; i < CH_COUNT; i++) {	// ADD 2009.12.15
				ws[i] = 0;					// ADD 2009.12.15
			}								// ADD 2009.12.15
		}

		//
		// Push switch operation
		//
		// Up bottun
		if (SW1) {
			if (sw1_scan == 0) {
				// Short push -> Roll up
				if (disp_mode == DISP_MODE_NORMAL) {
					if (++disp_ch >= CH_COUNT) {
						disp_ch = 0;
					}
				} else {
					disp_ee += sizeof(ee_buf);		// No overflow
				}

			} else if (sw1_scan > SW1_LONGPUSH) {
				// Long push -> DateTime mode
				dateset();
			}
			sw1_scan++;
		} else {
			sw1_scan = 0;
		}

		// Down bottun
		if (SW2) {
			// Short push -> Roll down
			if (sw2_scan == 0) {
				if (disp_mode == DISP_MODE_NORMAL) {
					if (disp_ch == 0) {
						disp_ch = CH_COUNT;
					}
					--disp_ch;
				} else {
					disp_ee -= sizeof(ee_buf);		// No overflow
				}

			// Long push -> History mode
			} else if (sw2_scan > SW2_LONGPUSH) {
				if (disp_mode == DISP_MODE_NORMAL) {
					disp_mode = DISP_MODE_HIST;
				} else {
					disp_mode = DISP_MODE_NORMAL;
				}
			}
			sw2_scan++;
		} else {
			sw2_scan = 0;
		}

		//
		// Display
		//
		if (!(ct & 0x3)) {
			if (disp_mode == DISP_MODE_NORMAL) {
//				int wh;					// DEL 2009.11.22
				static long wh;			// ADD 2009.11.22
				int kwh;
				int mod;
				int yen;

				wh = ws[disp_ch] / 3600;
				kwh = wh / 1000;
				mod = wh % 1000;
				yen = wh * YEN_RATE / 1000;

				LCD_Position(0,0);
				if (ct & 0x80) {
					printf("#%d%4dW %02d/%2d/%2d", disp_ch, w[disp_ch], yy, mm, dd);
				} else {
					printf("#%d%4dW %2d:%02d:%02d", disp_ch, w[disp_ch], hh, mi, ss);
				}
				LCD_Position(1,0);
//				if (wh < 10000) {		// DEL 2009.11.22
				if (kwh < 10) {			// ADD 2009.11.22
					printf("%2d.%03dkWh%6d\374", kwh, mod, yen);
				} else {
					printf("%4d.%1dkWh%6d\374", kwh, mod/100, yen);
				}
			} else {
				if (disp_ee_b != disp_ee) {
					disp_ee_b = disp_ee;
					eeprom_reads(disp_ee, (char *)&ee_buf, sizeof(ee_buf));
				}

				LCD_Position(0,0);
				printf("%4d%4d%4d%4d", ee_buf.wh[0], ee_buf.wh[1], ee_buf.wh[2], ee_buf.wh[3]);
				LCD_Position(1,0);
				printf("%4d%4d%4d%2d%2d", ee_buf.wh[4], ee_buf.wh[5], ee_buf.wh[6], ee_buf.yy, ee_buf.mm);
			}
		}

		//
		// Loop end
		//
		CLRWDT();
	}
}